var Engine;
/******/ (() => { // webpackBootstrap
/******/ 	"use strict";
/******/ 	var __webpack_modules__ = ({

/***/ "./index.js":
/*!******************!*\
  !*** ./index.js ***!
  \******************/
/***/ ((module, __unused_webpack_exports, __webpack_require__) => {



module.exports = __webpack_require__(/*! ./lib/Engine */ "./lib/Engine.js");


/***/ }),

/***/ "./lib/Context.js":
/*!************************!*\
  !*** ./lib/Context.js ***!
  \************************/
/***/ ((module) => {



module.exports = class Context {

	constructor(engine, parameters, rulesFired, currentRuleFlowActivated) {
		this.engine = engine;
		this.parameters = parameters || {};
		this.rulesFired = rulesFired || [];
		this._currentRuleFlowActivated = !! currentRuleFlowActivated;
	}

	initiateFlow(ruleFlow) {
		this._currentRuleFlowActivated = false;
	}

	endFlow() {
		this._currentRuleFlowActivated = true;
	}

	get currentRuleFlowActivated() {
		return this._currentRuleFlowActivated
	}

	ruleFired(rule) {
		this.rulesFired.push(rule);
		this._currentRuleFlowActivated = true;
	}

	/**
	 * Creates a new context bound to the new set of parameters
	 * @param {Object} newParameters
	 */
	bindParameters(newParameters) {
		const parameters = Object.assign({}, this.parameters, newParameters);
		return new Context(this.engine, parameters, this.rulesFired, this._currentRuleFlowActivated);
	}

}


/***/ }),

/***/ "./lib/Engine.js":
/*!***********************!*\
  !*** ./lib/Engine.js ***!
  \***********************/
/***/ ((module, __unused_webpack_exports, __webpack_require__) => {



const Context = __webpack_require__(/*! ./Context */ "./lib/Context.js"),
	ClosureRegistry = __webpack_require__(/*! ./closure/ClosureRegistry */ "./lib/closure/ClosureRegistry.js")

	//actions
	const action_error 		= __webpack_require__(/*! ./common/action/error */ "./lib/common/action/error.js");
	const action_identity 		= __webpack_require__(/*! ./common/action/identity */ "./lib/common/action/identity.js");
	const action_setResult 		= __webpack_require__(/*! ./common/action/setResult */ "./lib/common/action/setResult.js");
	const action_undefined 		= __webpack_require__(/*! ./common/action/undefined */ "./lib/common/action/undefined.js");

	//conditions
	const cond_always 		= __webpack_require__(/*! ./common/condition/always */ "./lib/common/condition/always.js");
	//const cond_dateRange 	= require("./common/condition/dateRange");
	const cond_default 		= __webpack_require__(/*! ./common/condition/default */ "./lib/common/condition/default.js");
	const cond_equal 		= __webpack_require__(/*! ./common/condition/equal */ "./lib/common/condition/equal.js");
	const cond_never 		= __webpack_require__(/*! ./common/condition/never */ "./lib/common/condition/never.js");
	const cond_random 		= __webpack_require__(/*! ./common/condition/random */ "./lib/common/condition/random.js");

	//transformer
	const transf_fixedValue 		= __webpack_require__(/*! ./common/transformer/fixedValue */ "./lib/common/transformer/fixedValue.js");
	const transf_get 		= __webpack_require__(/*! ./common/transformer/get */ "./lib/common/transformer/get.js");


module.exports = class Engine {

	constructor() {
		this.services = {};
		this.closures = new ClosureRegistry(this);		
		
		//Add default action Closures
		action_error(this);
		action_identity(this);
		action_setResult(this);
		action_undefined(this);
		
		//Add default conditions Closures
		cond_always(this);
		//cond_dateRange(this);
		cond_default(this);
		cond_equal(this);
		cond_never(this);
		cond_random(this);

		//Add default transformers Closures
		transf_fixedValue(this);
		transf_get(this);

	}

	add(definition, options) {
		const closureOrClosures = this.closures.parse(definition);

		// if I get an array, then I assume that it is an array of definitions, and add each of them
		if (Array.isArray(closureOrClosures)) {
			closureOrClosures.forEach(clos => this.closures.add(clos.name, clos, options))
		} else {
			// non-array case
			this.closures.add(closureOrClosures.name, closureOrClosures, options);
		}
	}

	reset() {
		this.closures = new ClosureRegistry(this);
	}

	process(closure, fact) {
		if (typeof (closure) === "string") {
			closure = this.closures.get(closure);
		}

		const context = new Context(this);
		try {
			return Promise.resolve(closure.process(fact, context)).then(fact => {
				context.fact = fact;
				return context;
			});
		} catch (error) {
			return Promise.reject(error);
		}
	}

	
	

}



/***/ }),

/***/ "./lib/closure/Closure.js":
/*!********************************!*\
  !*** ./lib/closure/Closure.js ***!
  \********************************/
/***/ ((module) => {



/**
 * A closure that its identified by a name. All closures expose the method
 * process(fact, context) that allows them to operate as predicates or
 * transfomers of a certain fact.
 *
 * @type {Closure}
 */
class Closure {

	constructor(name, options) {
		this.name = name;
		this.options = options || {};
	}

	get named() {
		return !!this.name;
	}

	/**
	 * Evaluates the closure against a certain fact
	 *
	 * @param {Object} fact								a fact
	 * @param {Context} context						an execution context.
	 * @param {Object} context.parameters the execution parameters, if any
	 * @param {Engine} context.engine			the rules engine
	 *
	 * @return {Object|Promise} the result or a promise of such result
	 */
	process(fact, context) {
		throw new Error("This is an abstract closure, how did you get to instantiate this?");
	}

	/**
	 * Binds this closure to a set of parameters. This will return a new Closure than
	 * when invoked it will ALWAYS pass the given parameters as a fields inside the
	 * context.parameters object.
	 *
	 * @param {String} name - the name, if specified, of the resulting bounded closure
	 * @param {Object} parameters - the parameters to bound to the closure
	 * @param {Engine} engine - the rules engine instance
	 */
	bind(name, parameters, engine) {
		// const missing = (this.options.required || []).find(required => parameters[required] === undefined)
		// if (missing) {
		// 	throw new Error(`Cannot instantiate provided closure '${this.name}'. Parameter ${missing} is unbounded`);
		// }

		// No need to perform any binding, there is nothing to bind
		if (! Object.keys(parameters).length) {
			return this;
		}

		// Replaces parameters that are set as closureParamters with actual closures!
		// TODO: do we really need this? can we do it differently? I hate expanding the options
		// list
		if (this.options.closureParameters) {
			this.options.closureParameters.forEach(parameter => {
				parameters[parameter] = engine.closures.parseOrValue(parameters[parameter]);
			})
		}

		return new BoundClosure(name, this, parameters)
	}

}

Closure.closureType = true;

/**
 * A closure bound to a certain set of parameters
 *
 * @type {BoundClosure}
 */
class BoundClosure extends Closure {

	constructor(name, closure, parameters) {
		super(name);
		this.closure = closure;
		this.parameters = parameters || {};
	}

	process(fact, context) {
		const newContext = context.bindParameters(this.parameters);
		return this.closure.process(fact, newContext);
	}

}

module.exports = Closure;


/***/ }),

/***/ "./lib/closure/ClosureReducer.js":
/*!***************************************!*\
  !*** ./lib/closure/ClosureReducer.js ***!
  \***************************************/
/***/ ((module, __unused_webpack_exports, __webpack_require__) => {



const Closure = __webpack_require__(/*! ./Closure */ "./lib/closure/Closure.js"),
	util = __webpack_require__(/*! ../util */ "./lib/util.js");

// Strategies define how we merge that every fact returns.

const reduceStrategies = {
	and: (prev, next) => prev && next,
	or: (prev, next) => prev || next,
	// This strategy requires the previous fact, while the others require the original to process conditions
	last: (_, next) => next,
};

/**
 * This is a closure composite that will reduce the fact execution through
 * a list of component closures. The result of each closure execution will
 * be used as fact for the next closure.
 *
 * @type {ClosureReducer}
 */
module.exports = class ClosureReducer extends Closure {
	constructor(name, closures, options) {
		super(name, options);
		this.closures = closures || util.raise(`Cannot build closure reducer [${name}] without closure chain`);
	}

	process(fact, context) {
		return this.reduce(0, fact, context);
	}

	reduce(index, fact, context) {
		if (this.closures.length <= index) {
			return fact;
		}

		return util.nowOrThen(this.closures[index].process(fact, context), newFact => {
			if (this.options.matchOnce && context.currentRuleFlowActivated) {
				return newFact;
			}
			if (this.options.strategy && this.options.strategy !== "last") {
				const reduceStrategy = reduceStrategies[this.options.strategy];
				return this.closures.length <= index + 1
					? newFact
					: reduceStrategy(newFact, this.reduce(index + 1, fact, context));
			}
			return reduceStrategies.last(newFact, this.reduce(index + 1, newFact, context));
		});
	}
};


/***/ }),

/***/ "./lib/closure/ClosureRegistry.js":
/*!****************************************!*\
  !*** ./lib/closure/ClosureRegistry.js ***!
  \****************************************/
/***/ ((module, __unused_webpack_exports, __webpack_require__) => {



const FunctionalClosure = __webpack_require__(/*! ./FunctionalClosure */ "./lib/closure/FunctionalClosure.js"),
	ClosureReducer = __webpack_require__(/*! ./ClosureReducer */ "./lib/closure/ClosureReducer.js"),
	RuleFlow = __webpack_require__(/*! ./RuleFlow */ "./lib/closure/RuleFlow.js"),
	Rule = __webpack_require__(/*! ./Rule */ "./lib/closure/Rule.js"),
	util = __webpack_require__(/*! ../util */ "./lib/util.js");

/**
 * ClosureFactory is the main entry point for creating closure out of the
 * closure definition json.
 *
 * It also acts as a registry for named-closure implementations which need to
 * be provided before parsing any named closure.
 *
 * @type {ClosureRegistry}
 */
module.exports = class ClosureRegistry {
	constructor(engine) {
		this.engine = engine;
		this.namedClosures = {};
	}
	/**
	 * Adds a closure to the registry by name hence becoming a
	 * NAMED closure.
	 *
	 * @param {String} name                 the name of the closure.
	 * @param {Closure|Function} closure    the Closure object or function implementation.
	 * @param {Object} options              closure registering options.
	 * @param {Boolean} options.override    this method will fail if a closure with
	 * 				the same name alrady exists, unless override is set to truthy.
	 *
	 * @return {Closure} the closure added
	 */
	add(name, closure, options = {}) {
		name || util.raise("Cannot add anonymous closure");

		if (closure.closureType) {
			options.required = options.required || closure.required;
			options.closureParameters = options.closureParameters || closure.closureParameters;
			closure = new closure(name, options);
		}
		if (typeof closure === "function") {
			closure = new FunctionalClosure(name, closure, options);
		}

		if (this.namedClosures[name] && !options.override) {
			throw new Error(`Already defined a closure with name '${name}'`);
		}
		this.namedClosures[name] = closure;
		return closure;
	}

	get(name) {
		const closure = this.namedClosures[name];
		if (!closure) {
			throw new Error(`Unexistent named closure [${name}]`);
		}
		return closure;
	}

	/**
	 * Creates a closure from its definition.
	 *
	 * If definition parameter is:
	 * - an array then a ClosureReducer will be created and each item in the array
	 * 	 will be parsed as a closure.
	 * - an object with the property `rules` then it's interpreted as a rule flow
	 * 	 (an special case of a ClosureReducer)
	 * - an object has either `when` or `then` properties it is assumed to be a Rule
	 * 	 and it is created parsing both `when` and `then` definition as closures.
	 *
	 * - if it is a string a parameterless implementation for it will be looked
	 *   up in the implementations registry.
	 * - if it is an object it will an implementation for `definition.closure`
	 *   will be looked up in the implementation registry.
	 *
	 * @param  {Object|string|Object[]} definition the json defintion for the closure
	 * @return {Object}            		a closure object (it will understand the
	 *                                message process)
	 */
	parse(definition, options) {
		if (Array.isArray(definition)) {
			return this._createReducer(definition, options); //closure reducer for arrays
		} else if (definition.rules) {
			return this._createRuleFlow(definition);
		} else if (definition.when || definition.then) {
			return this._createRule(definition);
		} else if (definition.closureLibrary) {
			return this._createClosureLibrary(definition);
		} else {
			return this._createNamedClosure(definition);
		}
	}

	parseOrValue(definition) {
		// if it is exactly undefined: do nothing
		if (definition === undefined) {
			return definition
		}

		// rule out the "value" case: it is a falsy value a number, an Array, or a String which is not the name of a registered closure
		// in such cases, I return in fact a fixedValue closure for the given value
		if (!definition || (typeof definition === "number") || Array.isArray(definition)
			|| ((typeof definition === "string") && !this.namedClosures[definition])
		) {
			return this.namedClosures["fixedValue"].bind(null, {value: definition}, null)  // no engine needed
		}

		// it is a true definition
		return this.parse(definition)
	}

	_createReducer(definition, options) {
		const closures = definition.map(eachDefinition => this.parse(eachDefinition));
		return new ClosureReducer(definition.name, closures, options);
	}

	_createRule(definition) {
		if (!definition.when) throw new Error(`Rule '${definition.name}' must define a valid when clause`);
		if (!definition.then) throw new Error(`Rule '${definition.name}' must define a valid then clause`);

		const condition = this.parse(definition.when, { strategy: definition.conditionStrategy || "and" });
		const action = this.parse(definition.then);
		return new Rule(definition.name, condition, action);
	}

	_createRuleFlow(definition) {
		const closures = definition.rules.map(eachDefinition => this.parse(eachDefinition));
		return new RuleFlow(definition.name, closures, { matchOnce: definition.matchOnce });
	}

	_createClosureLibrary(definition) {
		return definition.closureLibrary.map(closureDefinition => this._createNamedClosure(closureDefinition));
	}

	_createNamedClosure(definition) {
		definition = typeof definition === "string" ? { closure: definition } : definition;
		const closure = this.get(definition.closure);

		const parameters = Object.assign({}, definition);
		delete parameters.closure;

		return closure.bind(definition.name, parameters, this.engine);
	}
};


/***/ }),

/***/ "./lib/closure/FunctionalClosure.js":
/*!******************************************!*\
  !*** ./lib/closure/FunctionalClosure.js ***!
  \******************************************/
/***/ ((module, __unused_webpack_exports, __webpack_require__) => {



const Closure = __webpack_require__(/*! ./Closure */ "./lib/closure/Closure.js");

/**
 * A simple closure that's implemented through a function that is defined
 * in beforehand.
 *
 * @type {ProvidedClosure}
 */
class FunctionalClosure extends Closure {

	/**
	 * @param  {string}   name       the name of the clousre
	 * @param  {Function} fn         a function providing an implementation
	 *                               for this closure.
	 */
	constructor(name, fn, options) {
		super(name, options);
		if (typeof(fn) !== "function") {
			throw new TypeError(`Implementation for provided closure '${name}' is not a function`);
		}
		this.fn = fn;
	}

	/**
	 * Evaluates the block against a fact promise
	 * @param {Object} fact 						a fact
	 * @param {Context} context					an execution context.
	 * @param {Context} context.engine	the rules engine
	 *
	 * @return {Object|Promise} a promise that will be resolved to some result
	 */
	process(fact, context) {
		return this.fn.call(this, fact, context);
	}

}

module.exports = FunctionalClosure;


/***/ }),

/***/ "./lib/closure/Rule.js":
/*!*****************************!*\
  !*** ./lib/closure/Rule.js ***!
  \*****************************/
/***/ ((module, __unused_webpack_exports, __webpack_require__) => {



const Closure = __webpack_require__(/*! ./Closure */ "./lib/closure/Closure.js"),
	util = __webpack_require__(/*! ../util */ "./lib/util.js");

/**
 * A rule is a named conditional closure. It understands the `process` message
 * but it will only fire its internal closure action if the provided fact matches
 * the associated conditional closure.
 *
 * @type {Rule}
 */
module.exports = class Rule extends Closure{

	constructor(name, condition, action) {
		super(name);
		this.condition = condition || util.raise(`Cannot build rule [${name}] without condition closure`);
		this.action = action || util.raise(`Cannot build rule [${name}] without action closure`);
	}

	/**
	 * Executes the actions associated with this rule over certain fact
	 * @param {Object} fact 	   				a fact
	 * @param {Context} context 				an execution context
	 * @param {Context} context.engine	the rules engine
	 *
	 * @return {Object|Promise} a promise that will be resolved to some result (typically
	 *                   such result will be used as next's rule fact)
	 */
	process(fact, context) {
		return util.nowOrThen(this.evaluateCondition(fact, context), matches => {
			if (matches) {
				context.ruleFired(this);
				return this.action.process(fact, context);
			}
			return fact;
		});
	}

	/**
	 * Evaluates a condition
	 * @param  {Promise} fact 	a fact
	 * @param  {Context} engine	an execution context
	 * @return {Promise}				a Promise that will be resolved to a truthy/falsey
	 */
	evaluateCondition(fact, context) {
		return this.condition.process(fact, context);
	}

}


/***/ }),

/***/ "./lib/closure/RuleFlow.js":
/*!*********************************!*\
  !*** ./lib/closure/RuleFlow.js ***!
  \*********************************/
/***/ ((module, __unused_webpack_exports, __webpack_require__) => {



const ClosureReducer = __webpack_require__(/*! ./ClosureReducer */ "./lib/closure/ClosureReducer.js"),
	util = __webpack_require__(/*! ../util */ "./lib/util.js")

module.exports = class RuleFlow extends ClosureReducer {

	process(fact, context) {
		context.initiateFlow();
		return util.nowOrThen(super.process(fact, context), fact => {
			context.endFlow();
			return fact;
		});
	}

}


/***/ }),

/***/ "./lib/common/action/error.js":
/*!************************************!*\
  !*** ./lib/common/action/error.js ***!
  \************************************/
/***/ ((module) => {



module.exports = engine => {
	engine.closures.add("error", (fact, {parameters}) => { throw new Error(parameters.message) }, {
		required: ["message"]
	});
}


/***/ }),

/***/ "./lib/common/action/identity.js":
/*!***************************************!*\
  !*** ./lib/common/action/identity.js ***!
  \***************************************/
/***/ ((module) => {



module.exports = engine => {
	engine.closures.add("identity", fact => fact);
}


/***/ }),

/***/ "./lib/common/action/setResult.js":
/*!****************************************!*\
  !*** ./lib/common/action/setResult.js ***!
  \****************************************/
/***/ ((module, __unused_webpack_exports, __webpack_require__) => {



const util = __webpack_require__(/*! ../../util */ "./lib/util.js")

module.exports = engine => {
	const fn = (fact, context)  => {
		const result = context.parameters.calculator.process(fact, context);
		return util.nowOrThen(result, value => {
			fact[context.parameters.field] = value;
			return fact;
		});
	}

	engine.closures.add("setResult", fn, {
		required: ["field", "calculator"],
		closureParameters: [ "calculator" ]
	});
}


/***/ }),

/***/ "./lib/common/action/undefined.js":
/*!****************************************!*\
  !*** ./lib/common/action/undefined.js ***!
  \****************************************/
/***/ ((module) => {



module.exports = engine => {
	engine.closures.add("undefined", fact => undefined);
}


/***/ }),

/***/ "./lib/common/condition/always.js":
/*!****************************************!*\
  !*** ./lib/common/condition/always.js ***!
  \****************************************/
/***/ ((module) => {



module.exports = engine => {
	engine.closures.add("always", () => true);
}


/***/ }),

/***/ "./lib/common/condition/default.js":
/*!*****************************************!*\
  !*** ./lib/common/condition/default.js ***!
  \*****************************************/
/***/ ((module) => {



module.exports = engine => {
	//Executes only if NO other rule has been executed for this flow
	engine.closures.add("default", (fact, context) => ! context.currentRuleFlowActivated );
}


/***/ }),

/***/ "./lib/common/condition/equal.js":
/*!***************************************!*\
  !*** ./lib/common/condition/equal.js ***!
  \***************************************/
/***/ ((module) => {



module.exports = engine => {
	//Matches only if certain fact field (specified by field) equals to some
	//other specified value
	const fn = (fact, context) => {
		const factValue = context.parameters.field ? fact[context.parameters.field] : fact;
		return factValue === context.parameters.value
	};

	engine.closures.add("equal", fn, {
		required: ["value"],
		optionalParameters: ["field"]
	});
}


/***/ }),

/***/ "./lib/common/condition/never.js":
/*!***************************************!*\
  !*** ./lib/common/condition/never.js ***!
  \***************************************/
/***/ ((module) => {



module.exports = engine => {
	engine.closures.add("never", () => false);
}


/***/ }),

/***/ "./lib/common/condition/random.js":
/*!****************************************!*\
  !*** ./lib/common/condition/random.js ***!
  \****************************************/
/***/ ((module) => {



module.exports = engine => {
	engine.closures.add("random", () => Math.random() >= 0.5);
}


/***/ }),

/***/ "./lib/common/transformer/fixedValue.js":
/*!**********************************************!*\
  !*** ./lib/common/transformer/fixedValue.js ***!
  \**********************************************/
/***/ ((module) => {



module.exports = engine => {
	const fn = (fact, context)  => { return context.parameters.value }

	engine.closures.add("fixedValue", fn, { required: ["value"] });
}


/***/ }),

/***/ "./lib/common/transformer/get.js":
/*!***************************************!*\
  !*** ./lib/common/transformer/get.js ***!
  \***************************************/
/***/ ((module) => {



module.exports = engine => {
	const fn = (fact, context)  => { return fact[context.parameters.prop] }

	engine.closures.add("get", fn, { required: ["prop"] });
}


/***/ }),

/***/ "./lib/util.js":
/*!*********************!*\
  !*** ./lib/util.js ***!
  \*********************/
/***/ ((module) => {



module.exports = {

	raise(message) {
		throw new Error(message);
	},

	nowOrThen(p, block) {
		if (p && p.then) {
			return p.then(block);
		} else {
			return block(p)
		}
	}

}


/***/ })

/******/ 	});
/************************************************************************/
/******/ 	// The module cache
/******/ 	var __webpack_module_cache__ = {};
/******/ 	
/******/ 	// The require function
/******/ 	function __webpack_require__(moduleId) {
/******/ 		// Check if module is in cache
/******/ 		var cachedModule = __webpack_module_cache__[moduleId];
/******/ 		if (cachedModule !== undefined) {
/******/ 			return cachedModule.exports;
/******/ 		}
/******/ 		// Create a new module (and put it into the cache)
/******/ 		var module = __webpack_module_cache__[moduleId] = {
/******/ 			// no module.id needed
/******/ 			// no module.loaded needed
/******/ 			exports: {}
/******/ 		};
/******/ 	
/******/ 		// Execute the module function
/******/ 		__webpack_modules__[moduleId](module, module.exports, __webpack_require__);
/******/ 	
/******/ 		// Return the exports of the module
/******/ 		return module.exports;
/******/ 	}
/******/ 	
/************************************************************************/
/******/ 	
/******/ 	// startup
/******/ 	// Load entry module and return exports
/******/ 	// This entry module used 'module' so it can't be inlined
/******/ 	var __webpack_exports__ = __webpack_require__("./index.js");
/******/ 	Engine = __webpack_exports__;
/******/ 	
/******/ })()
;
//# sourceMappingURL=rules-js.js.map